/***************************************************************************
 *   Copyright (C) 2008~2010 by wind (xihe)                                *
 *   xihels@gmail.com                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   51 Franklin St, Fifth Floor, Boston, MA 02110-1301, USA.              *
 ***************************************************************************/
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <unistd.h>
#include <libintl.h>
#include <fcntl.h>
#include <limits.h>
#include <sys/time.h>

#include "fcitx/fcitx.h"
#include "fcitx-utils/log.h"
#include "fcitx/module.h"
#include "fcitx/frontend.h"
#include "fcitx/instance.h"
#include "fcitx/candidate.h"
#include "fcitx-utils/utils.h"
#include "fcitx-utils/utils.h"

#define MAX_IMNAME_LEN 30
#define READ_TIMEOUT 5000 /* ms */

static void* RemoteCreate(FcitxInstance* instance);
static void RemoteProcessEvent(void* arg);
static void RemoteSetFD(void* arg);
static void RemoteDestroy(void* arg);
static int CreateSocket(const char *name);

FCITX_DEFINE_PLUGIN(fcitx_remote, module, FcitxModule) = {
    RemoteCreate,
    RemoteSetFD,
    RemoteProcessEvent,
    RemoteDestroy,
    NULL
};

typedef struct _FcitxRemote {
    FcitxInstance* owner;
    int socket_fd;
} FcitxRemote;

static char commit_str[MAX_USER_INPUT * UTF8_MAX_LENGTH + 1];

void* RemoteCreate(FcitxInstance* instance)
{
    FcitxRemote* remote = fcitx_utils_malloc0(sizeof(FcitxRemote));
    remote->owner = instance;

    char *socketfile;
    asprintf(&socketfile, "/tmp/fcitx-socket-:%d", fcitx_utils_get_display_number());
    remote->socket_fd = CreateSocket(socketfile);
    if (remote->socket_fd < 0) {
        FcitxLog(ERROR, _("Can't open socket %s: %s"), socketfile, strerror(errno));
        free(remote);
        free(socketfile);
        return NULL;
    }

    fcntl(remote->socket_fd, F_SETFD, FD_CLOEXEC);
    fcntl(remote->socket_fd, F_SETFL, O_NONBLOCK);
    chmod(socketfile, 0600);
    free(socketfile);
    return remote;
}

int CreateSocket(const char *name)
{
    int fd;
    int r;
    struct sockaddr_un uds_addr;

    /* JIC */
    unlink(name);

    fd = socket(AF_UNIX, SOCK_STREAM, 0);
    if (fd < 0) {
        return fd;
    }

    int opt = 1;
    setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, (void*) &opt, sizeof(opt));

    /* setup address struct */
    memset(&uds_addr, 0, sizeof(uds_addr));
    uds_addr.sun_family = AF_UNIX;
    strcpy(uds_addr.sun_path, name);

    /* bind it to the socket */
    r = bind(fd, (struct sockaddr *)&uds_addr, sizeof(uds_addr));
    if (r < 0) {
        return r;
    }


    /* listen - allow 10 to queue */
    r = listen(fd, 10);
    if (r < 0) {
        return r;
    }

    return fd;
}


int UdAccept(int listenfd)
{
    for (int i =0;i<20;i++) {
        int newsock = 0;
        struct sockaddr_un cliaddr;
        socklen_t len = sizeof(struct sockaddr_un);

        newsock = accept(listenfd, (struct sockaddr *)&cliaddr, &len);
        if (newsock < 0) {
            if (errno == EINTR) {
                continue; /* signal */
            }
        }

        return newsock;
    }
    return -1;
}

static void SendIMState(FcitxRemote* remote, int fd)
{
    FcitxContextState r = FcitxInstanceGetCurrentState(remote->owner);
    write(fd, &r, sizeof(r));
}

FCITX_EXPORT_API
void RemoteCommitString(const char *str)
{
    if (str && ((strlen(commit_str) + strlen(str)) < sizeof(commit_str)))
        strcat(commit_str, str);
}

FCITX_EXPORT_API
char * RemoteGetCommitString()
{
    return commit_str;
}

FCITX_EXPORT_API
void RemoteClearCommitString()
{
    *commit_str = '\0';
}

static ssize_t readn(int fd, void *vptr, size_t n, unsigned timeout)
{
    size_t  nleft;
    ssize_t nread;
    char    *ptr;
    struct timeval current_time;
    uint64_t start_time;
    uint64_t cur_time;

    gettimeofday(&current_time, NULL);
    cur_time = (current_time.tv_sec * 1000LL) + (current_time.tv_usec / 1000LL);
    start_time = cur_time;

    ptr = vptr;
    nleft = n;
    while (1) {
        if ((nread = read(fd, ptr, nleft)) < 0) {
            if (errno != EWOULDBLOCK)
                return -1;
            nread = 0;
        } else if (nread == 0)
            break;

        nleft -= nread;
        ptr   += nread;

        if (nleft <= 0)
            break;

        gettimeofday(&current_time, NULL);
        cur_time = (current_time.tv_sec * 1000LL) + (current_time.tv_usec / 1000LL);
        if ((cur_time - start_time) > timeout)
            break;

        usleep(1000);
    }
    return (n - nleft);
}

static int SetInputMethod(FcitxInstance* instance, const char* name)
{
    if(!FcitxInstanceGetIMFromIMList(instance, IMAS_Enable, name)) {
        char* im_list;
        FcitxProfile* profile;

        if(!FcitxInstanceGetIMFromIMList(instance, IMAS_Disable, name)) {
            return -1;
        }

        profile = FcitxInstanceGetProfile(instance);
        if (profile->imList)
            fcitx_utils_alloc_cat_str(im_list, profile->imList, ",", name, ":", "True"); // TODO: replace instead add
        else
            fcitx_utils_alloc_cat_str(im_list, name, ":", "True");

        if (profile->imList)
            free(profile->imList);
        profile->imList = im_list;

        FcitxInstanceUpdateIMList(instance);
    }

    FcitxInstanceSwitchIMByNameDoNotSkipZero(instance, name);

    return 0;
}
static int HandleSetInputMethodCmd(int fd, FcitxInstance* instance)
{
    int im_name_len;

    /* read im name */

    if (readn(fd, &im_name_len, sizeof(im_name_len), READ_TIMEOUT) != sizeof(im_name_len)) {
        FcitxLog(ERROR, "Can't read im name string length.");
        printf("Can not read im name string length.\n");
        return -1;
    }

    if (im_name_len) {
        char *im_name = malloc(im_name_len);
        if (!im_name) {
            FcitxLog(ERROR, "Can't allocate mem for im name string.");
            printf("Can not allocate mem for im name string.\n");
            return -1;
        }

        if (readn(fd, im_name, im_name_len, READ_TIMEOUT) != im_name_len) {
            FcitxLog(ERROR, "Can't read im name.");
            printf("Can not read im name.\n");
            free(im_name);
            return -1;
        }

        /* set im */

        if (SetInputMethod(instance, im_name)) {
            FcitxLog(ERROR, "Can't set im.");
            printf("Can not set im.\n");
            free(im_name);
            return -1;
        }

        free(im_name);
    }

    return 0;
}

static int ProcessKey(FcitxInstance* instance, FcitxInputContext* callic, const uint32_t originsym, const uint32_t keycode, const uint32_t originstate, uint32_t t, FcitxKeyEventType type)
{
    FcitxInputContext* ic = FcitxInstanceGetCurrentIC(instance);
    FcitxGlobalConfig* config = FcitxInstanceGetGlobalConfig(instance);
    FcitxInputState* input = FcitxInstanceGetInputState(instance);

    if (ic == NULL || ic->frontendid != callic->frontendid) {
        FcitxInstanceSetCurrentIC(instance, callic);
        FcitxUIOnInputFocus(instance);
    }
    ic = callic;
    FcitxKeySym sym;
    unsigned int state;

    state = originstate & FcitxKeyState_SimpleMask;
    state &= FcitxKeyState_UsedMask;
    FcitxHotkeyGetKey(originsym, state, &sym, &state);
    FcitxLog(DEBUG,
             "KeyRelease=%d  state=%d  KEYCODE=%d  KEYSYM=%u ",
             (type == FCITX_RELEASE_KEY), state, keycode, sym);

    if (originsym == 0)
        return 0;

    if (ic->state == IS_CLOSED && type == FCITX_PRESS_KEY && FcitxHotkeyIsHotKey(sym, state, config->hkTrigger)) {
        FcitxInstanceEnableIM(instance, ic, false);
        return 1;
    }
    else if (ic->state == IS_CLOSED) {
        return 0;
    }

    FcitxInputStateSetKeyCode(input, keycode);
    FcitxInputStateSetKeySym(input, originsym);
    FcitxInputStateSetKeyState(input, originstate);
    INPUT_RETURN_VALUE retVal = FcitxInstanceProcessKey(instance, type,
                                           t,
                                           sym, state);
    FcitxInputStateSetKeyCode(input, 0);
    FcitxInputStateSetKeySym(input, 0);
    FcitxInputStateSetKeyState(input, 0);

    if (retVal & IRV_FLAG_FORWARD_KEY || retVal == IRV_TO_PROCESS)
        return 0;
    else
        return 1;
}

static int FcitxGetCandidateWordsMsg(FcitxInstance* instance, char **msg, int *len)
{
    FcitxInputState *input = FcitxInstanceGetInputState(instance);
    FcitxCandidateWordList *cand_list = FcitxInputStateGetCandidateList(input);
    FcitxCandidateWord* cand_word;
    char *msg_pos;
    int msg_len = 0;

    /* calc msg str len */

    if (FcitxCandidateWordGetFirst(cand_list)) {
        for (cand_word = FcitxCandidateWordGetFirst(cand_list);
             cand_word != NULL;
             cand_word = FcitxCandidateWordGetNext(cand_list, cand_word)) {
            if (cand_word->strWord)
                msg_len += strlen(cand_word->strWord) +1; /* +1 for '\0' */
            else
                msg_len++; /* +1 for '\0' */

            if (cand_word->strExtra)
                msg_len += strlen(cand_word->strExtra) +1; /* +1 for '\0' */
            else
                msg_len++; /* for '\0' */
        }
    } else {
        char *commit_str = RemoteGetCommitString();
        if (commit_str)
            msg_len += strlen(commit_str);

        char* strPreedit = FcitxUIMessagesToCString(FcitxInputStateGetClientPreedit(input));
        if (strPreedit) {
            msg_len += strlen(strPreedit);
            free(strPreedit);
        }

        msg_len += 2; /* +2 for word and extra '\0' */
    }

    /* ... */

    *msg = malloc(msg_len);
    if(!*msg)
        return -1;

    /* cpy str */

    msg_pos = *msg;

    if (FcitxCandidateWordGetFirst(cand_list)) {
        for (cand_word = FcitxCandidateWordGetFirst(cand_list);
             cand_word != NULL;
             cand_word = FcitxCandidateWordGetNext(cand_list, cand_word)) {
            if (cand_word->strWord)
                msg_pos = stpcpy(msg_pos, cand_word->strWord) +1; /* +1 for '\0' */
            else
                *msg_pos++ = '\0';

            if (cand_word->strExtra)
                msg_pos = stpcpy(msg_pos, cand_word->strExtra) +1; /* +1 for '\0' */
            else
                *msg_pos++ = '\0';
        }
    } else {
        char *commit_str = RemoteGetCommitString();
        if (commit_str)
            msg_pos = stpcpy(msg_pos, commit_str);

        char* strPreedit = FcitxUIMessagesToCString(FcitxInputStateGetClientPreedit(input));
        if (strPreedit) {
            msg_pos = stpcpy(msg_pos, strPreedit);
            free(strPreedit);
        }

        if (!strlen(*msg))
            *msg_pos++ = '\0'; /* word */

        *msg_pos++ = '\0'; /* extra */
    }

    *len = msg_len;
    return 0;
}

static int HandleProcessKeyCmd(int fd, FcitxInstance* instance)
{
    uint32_t keycode = 0;
    uint32_t sym;
    uint32_t state = 0;
    uint32_t time = 0;
    int len;
    FcitxInputContext* ic;
    char *candidate_words;

    /* read symbol */

    if (readn(fd, &sym, sizeof(sym), READ_TIMEOUT) != sizeof(sym)) {
        FcitxLog(ERROR, "Can't read sym.");
        return -1;
    }

    /* handle symbol */

    ic = FcitxInstanceGetCurrentIC(instance);
    if (!ic) {
        ic = FcitxInstanceCreateIC(instance, 0, NULL);
        if (!ic) {
            FcitxLog(ERROR, "Can't create IC.");
            return -1;
        }
        FcitxInstanceSetCurrentIC(instance, ic);
    }

    ProcessKey(instance, ic, sym, keycode, state, time, FCITX_PRESS_KEY); // TODO: ret code
    ProcessKey(instance, ic, sym, keycode, state, time, FCITX_RELEASE_KEY);

    /* get results */

    if (FcitxGetCandidateWordsMsg(instance, &candidate_words, &len) == -1) {
        FcitxLog(ERROR, "Can't get candidate words.");
        return -1;
    }

    /* write results */

    if (write(fd, &len, sizeof(len)) != sizeof(len)) {
        FcitxLog(ERROR, "Can't write candidate words message length.");
        if (len)
            free(candidate_words);
        return -1;
    }

    if (len) {
        if (write(fd, candidate_words, len) != len) {
            FcitxLog(ERROR, "Can't write candidate words message.");
            free(candidate_words);
            return -1;
        }
        free(candidate_words);
    }

    return 0;
}

static int HandleTextEnteredCmd(FcitxInstance* instance)
{
    uint32_t keycode = 0;
    uint32_t sym = 0x20; /* space code */
    uint32_t state = 0;
    uint32_t time = 0;
    FcitxInputContext* ic;

    ic = FcitxInstanceGetCurrentIC(instance);
    if (!ic) {
        ic = FcitxInstanceCreateIC(instance, 0, NULL);
        if (!ic) {
            FcitxLog(ERROR, "Can't create IC.");
            return -1;
        }
        FcitxInstanceSetCurrentIC(instance, ic);
    }

    ProcessKey(instance, ic, sym, keycode, state, time, FCITX_PRESS_KEY);
    ProcessKey(instance, ic, sym, keycode, state, time, FCITX_RELEASE_KEY);

    RemoteClearCommitString();

    return 0;
}

static void RemoteProcessEvent(void* p)
{
    FcitxRemote* remote = (FcitxRemote*) p;
    unsigned int O;
    // lower 16bit 0->get 1->set, 2->reload, 3->toggle, 4->setIM
    int client_fd = UdAccept(remote->socket_fd);
    if (client_fd < 0)
        return;
    read(client_fd, &O, sizeof(int));
    unsigned int cmd = O & 0xFFFF;
    unsigned int arg = (O >> 16) & 0xFFFF;
    switch (cmd) {
        /// {{{
    case 0:
        SendIMState(remote, client_fd);
        break;
    case 1:
        /* arg is not state, due to backward compatible */
        if (arg == 0)
            FcitxInstanceCloseIM(remote->owner, FcitxInstanceGetCurrentIC(remote->owner));
        else {
            FcitxInstanceEnableIM(remote->owner, FcitxInstanceGetCurrentIC(remote->owner), false);
        }
        break;
    case 2:
        FcitxInstanceReloadConfig(remote->owner);
        break;
    case 3:
        FcitxInstanceChangeIMState(remote->owner, FcitxInstanceGetCurrentIC(remote->owner));
        break;
    case 4: {
        char imname[MAX_IMNAME_LEN];
        int n = read(client_fd, imname, MAX_IMNAME_LEN-1);
        imname[n] = '\0';
        FcitxInstanceSwitchIMByName(remote->owner, imname);
        break;
    }
    case 5: {
        int val;
        if ((val = fcntl(client_fd, F_GETFL, 0)) < 0)
            break;
        if (fcntl(client_fd, F_SETFL, val | O_NONBLOCK) < 0)
            break;

        (void)HandleSetInputMethodCmd(client_fd, remote->owner);

        break;
    }
    case 6: {
        int val;
        if ((val = fcntl(client_fd, F_GETFL, 0)) < 0)
            break;
        if (fcntl(client_fd, F_SETFL, val | O_NONBLOCK) < 0)
            break;

        (void)HandleProcessKeyCmd(client_fd, remote->owner);

        break;
    }
    case 7:
        (void)HandleTextEnteredCmd(remote->owner);
        break;
    default:
        break;
        /// }}}
    }
    close(client_fd);
}

void RemoteSetFD(void* arg)
{
    FcitxRemote* remote = (FcitxRemote*) arg;
    FD_SET(remote->socket_fd, FcitxInstanceGetReadFDSet(remote->owner));
    if (FcitxInstanceGetMaxFD(remote->owner) < remote->socket_fd)
        FcitxInstanceSetMaxFD(remote->owner, remote->socket_fd);
}

void RemoteDestroy(void* arg)
{
    FcitxRemote* remote = (FcitxRemote*) arg;
    close(remote->socket_fd);
    free(remote);
}


// kate: indent-mode cstyle; space-indent on; indent-width 0;
